#!/usr/local/bin/perl -w

=head1 NAME

cvs.t - testing of vcp cvs i/o

=cut

use strict ;

use Carp ;
use Cwd ;
use File::Spec ;
use Test ;

use VCP::TestUtils ;

# use eq_or_diff from Test::Differences if available,
# else just use ok.
eval "use Test::Differences";
sub ok_or_diff {
   goto &eq_or_diff if defined &eq_or_diff;
   goto &Test::ok;
}

my @vcp = vcp_cmd ;

my $t = -d 't' ? 't/' : '' ;

my $module = 'foo' ;  ## Must match the rev_root in the testrevml files

my @revml_out_spec = ( "revml:", "--sort-by=name,rev_id" ) ;

my $max_change_id ;

my $p4_spec_base ;
my $p4d_borken = $ENV{P4DBORKEN} || p4d_borken ;

sub check {
   croak "test failed and FATALTEST set" if ! $_[0];
}


my $infile_0 = $t . "test-cvs-in-0.revml";
my $cvsroot_0 = $t . "cvsroot_0";
my $infile_1 = $t . "test-cvs-in-1.revml";
my $cvsroot_1 = $t . "cvsroot_1";

my $cvs_spec_0        = "cvs:$cvsroot_0:$module/" ;
my $cvs_spec_1        = "cvs:$cvsroot_1:$module/" ;

my @tests = (
sub {
   return skip $p4d_borken, 1, 1, $p4d_borken if $p4d_borken ;

   $ENV{P4USER}   = "foobar_user" ;
   $ENV{P4PORT}   = "foobar_port" ;
   $ENV{P4CLIENT} = "foobar_client" ;
   $ENV{P4PASSWD} = "foobar_passwd" ;

   my $p4_options = launch_p4d "cvs_" ;
   $p4_spec_base = "p4:$p4_options->{user}:\@$p4_options->{port}:" ;
   check ok 1 ;
},

##
## cvs->p4->revml
##
sub {
   return skip $p4d_borken, 1, 1, $p4d_borken if $p4d_borken ;

   my $cvs_spec = $cvs_spec_0;
   my $p4_spec = "$p4_spec_base//depot" ;

   my $out;
   eval { run [ @vcp, "vcp:-" ], \<<TOHERE; };
Source:      $cvs_spec_0
Destination: $p4_spec_base//depot/
Map:
	(...)<>         main/\$1  # Match files on main branch (only)
        (...)<(*)>      \$2/\$1    # Match files on branches
TOHERE
   die $@ if $@;

   eval {
      run [ @vcp, "$p4_spec/...", @revml_out_spec ], \undef, \$out;
   };
   die $@ if $@;

   my $infile  = $t . "test-p4-in-0.revml";
   my $in = slurp $infile ;

   s_content  qw( rep_desc time user_id rep_type ), \$in, \$out ;
   s_content  qw( rev_root                       ), \$in, "depot" ;
   rm_elts    qw( p4_info                        ), \$in, \$out ;
   rm_elts    qw( label ), qr/r_\w+|ch_\w+/,              \$out ;

# We need to get VCP::Dest::p4 to build up branch specifications.  Until
# then, don't test branches.
rm_elts    qw( branches ), \$in, \$out ;
$in =~ s{(branch_id>)main-branch-1}{$1__branch_1}g;

   $in =~ s{(id="|id>)/ignored}{$1//depot}g;
   assert_eq $infile, $in, $out ;
   for ( $out =~ m{<change_id>(.*?)</change_id>}g ) {
      $max_change_id = $1
         if ! defined $max_change_id || $1 > $max_change_id ;
   }

   ok_or_diff $out, $in;
},

sub {
   return skip $p4d_borken, 1, 1, $p4d_borken if $p4d_borken ;
   check ok $max_change_id, 3, "Max change_id in cvs->p4 transfer" ;
},

##
## cvs->p4->revml, re-rooting a dir tree
##
sub {
   return skip $p4d_borken, 1, 1, $p4d_borken if $p4d_borken ;

   ## Hide global $cvs_spec for the nonce
   my $cvs_spec = "$cvs_spec_0/a/deeply/..." ;

   my $p4_spec = "$p4_spec_base//depot/new/..." ;

   my $out ;
   run [ @vcp, $cvs_spec, $p4_spec ], \undef;
   run [ @vcp, $p4_spec, @revml_out_spec ], \undef, \$out;

   my $infile  = $t . "test-p4-in-0.revml";
   my $in = slurp $infile ;

   s_content  qw( rep_desc time user_id rep_type  ), \$in, \$out ;
   s_content  qw( rev_root                        ), \$in, "depot/new" ;
   rm_elts    qw( p4_info                         ), \$in, \$out ;
   rm_elts    qw( label ),  qr/r_\w+|ch_\w+/,              \$out ;

   ## Strip out all files from $in that shouldn't be there
   rm_elts    qw( rev ), qr{(?:(?!a/deeply).)*?}s, \$in ;

   ## Adjust the $in paths to look like the result paths.  $in is
   ## now the "expected" output.
   $in =~ s{<(name|source_name)>main/a/deeply/}{<$1>}g ;

   for ( $out =~ m{<change_id>(.*?)</change_id>}g ) {
      $max_change_id = $1
         if ! defined $max_change_id || $1 > $max_change_id ;
   }

   $in =~ s{ (< (?:source_)?change_id >) (.*?) (</ (?:source_)?change_id >) }
           { $1 . ( $2 + 3 ) . $3 }gex;

   $in =~ s{(id="|id>)/ignored/main/a/deeply}{$1//depot/new}g;  #"# cperl syntax highlight fix

# We need to get VCP::Dest::p4 to build up branch specifications.  Until
# then, don't test branches.
rm_elts    qw( branches ), \$in, \$out ;
$in =~ s{(branch_id>)main-branch-1}{$1__branch_1}g;

   ok_or_diff $out, $in;
},

sub {
   return skip $p4d_borken, 1, 1, $p4d_borken if $p4d_borken ;
   check ok $max_change_id, 6, "Max change_id in cvs->p4 transfer" ;
},

##
## Incremental cvs->p4->revml update
##
sub {
   return skip $p4d_borken, 1, 1, $p4d_borken if $p4d_borken ;

   my $p4_spec = "$p4_spec_base//depot" ;

   my $out ;
   my $first_change = $max_change_id + 1 ;
   
   eval { run [ @vcp, "vcp:-" ], \<<TOHERE; };
Source:      $cvs_spec_1 -r ch_4:

Destination: $p4_spec_base//depot/...

Map:
	(...)<>         main/\$1  # Match files on main branch (only)
        (...)<(*)>      \$2/\$1    # Match files on branches
TOHERE
   die $@ if $@;

   eval {
      run [ @vcp, "$p4_spec/...\@$first_change,#head", @revml_out_spec ], \undef, \$out;
   };
   die $@ if $@;
   
   my $infile  = $t . "test-p4-in-1.revml";
   my $in = slurp $infile ;
   
   s_content  qw( rep_desc time user_id rep_type ), \$in, \$out ;
   s_content  qw( rev_root                       ), \$in, "depot" ;
   rm_elts    qw( p4_info                        ), \$in, \$out ;
   rm_elts    qw( label ), qr/r_\w+|ch_\w+/,              \$out ;
   
   for ( $out =~ m{<change_id>(.*?)</change_id>}g ) {
      $max_change_id = $1
         if ! defined $max_change_id || $1 > $max_change_id ;
   }
   
   $in =~ s{ (< (?:source_)?change_id >) ([4-9]|\d\d+) (</ (?:source_)?change_id >) }
           { $1 . ( $2 + 3 ) . $3 }gex;

#   $in =~ s{(<change_id>)([4-9]|\d\d+)(</change_id>)}{
#      $1 . ( $2 + 3 ) . $3;
#   }ge;
   
   $in =~ s{(id="|id>)/ignored}{$1//depot}g;
   
# We need to get VCP::Dest::p4 to build up branch specifications.  Until
# then, don't test branches.
rm_elts    qw( branches ), \$in, \$out ;
$in =~ s{(branch_id>)main-branch-1}{$1__branch_1}g;

   # the ch_4 is on the main trunk in CVS, not on the branch, so
   # these just won't be there.
   # TODO: enable multiple -r specs in VCP::Source::cvs or come
   # up with a state file that indicates what's been transferred.
   $in =~ s{^\s*<rev id="[^"]+main-branch-1[^"]+">.*?</rev>\n}{}gsm;

   ok_or_diff $out, $in;
},

sub {
   return skip $p4d_borken, 1, 1, $p4d_borken if $p4d_borken ;
   check ok $max_change_id, 9, "Max change_id in cvs->p4 transfer" ;
},

) ;

plan tests => scalar( @tests ) ;

my $why_skip ;

$why_skip .= cvs_borken ;
$why_skip ? skip( $why_skip, 0 ) : $_->() for @tests ;
